use std::fs;
use std::io::{self, Read};
use std::path::Path;

use color_eyre::eyre::{self, Context};
use color_eyre::Result;
use reqwest::blocking::{Client, Response};
use reqwest::header::HeaderMap;
use serde::{Deserialize, Serialize};
use time::OffsetDateTime;
use url::Url;

use crate::types::PullRequest;
use crate::Action;

static USER_AGENT: &str = concat!(env!("CARGO_PKG_NAME"), "/", env!("CARGO_PKG_VERSION"));

#[derive(Deserialize, Serialize, Debug)]
struct Source {
    access_token: String,
    owner: String,
    repo: String,
    url: Url,
}

#[derive(Deserialize, Serialize, Debug)]
struct Version {
    prs: Vec<String>,
    #[serde(with = "time::serde::timestamp")]
    timestamp: OffsetDateTime,
}

#[derive(Deserialize, Serialize, Debug)]
struct Config {
    /// Resource configuration.
    /// Passed verbatim from the definition in the pipeline.
    source: Source,
    /// For 'check':
    /// Previous version of this resource.
    /// Will be empty for the very first request.
    /// For 'in':
    /// The version metadata to create this as.
    version: Option<Version>,
}

#[tracing::instrument(skip_all)]
fn make_client(src: &Source) -> Result<Client> {
    let mut headers = HeaderMap::new();
    headers.insert(
        "Authorization",
        format!("token {}", src.access_token).try_into()?,
    );

    Client::builder()
        .default_headers(headers)
        .user_agent(USER_AGENT)
        .build()
        .map_err(From::from)
}

#[tracing::instrument(skip_all)]
fn fetch(src: &Source) -> Result<Response> {
    let client = make_client(&src).wrap_err("Failed to create HTTP client")?;
    let url = src
        .url
        .join(&format!("api/v1/repos/{}/{}/pulls", src.owner, src.repo))
        .wrap_err("Invalid URL")?;

    client
        .get(url)
        .query(&[("sort", "oldest"), ("state", "open")])
        .send()
        .map_err(From::from)
}

#[tracing::instrument(skip_all)]
fn action_check(conf: Config) -> Result<()> {
    let prs: Vec<PullRequest> = fetch(&conf.source)?.json()?;
    let mut version = Version {
        prs: prs.iter().map(|pr| pr.number.to_string()).collect(),
        timestamp: OffsetDateTime::now_utc(),
    };
    version.prs.sort_unstable();

    let out = if let Some(prev) = conf.version {
        if prev.prs == version.prs {
            vec![prev]
        } else {
            vec![prev, version]
        }
    } else {
        vec![version]
    };

    serde_json::to_writer_pretty(io::stdout(), &out)
        .wrap_err("Failed to write result to stdout")?;

    Ok(())
}

#[tracing::instrument(skip_all)]
fn action_in(conf: Config, dest: impl AsRef<Path>) -> Result<()> {
    let version = if let Some(version) = conf.version {
        version
    } else {
        eyre::bail!("Version missing in 'in' action.");
    };

    let bytes = fetch(&conf.source)?.bytes()?;
    let prs: Vec<PullRequest> = serde_json::from_slice(&bytes)?;

    {
        let mut prs: Vec<_> = prs.iter().map(|pr| pr.number.to_string()).collect();
        prs.sort_unstable();
        if version.prs != prs {
            eyre::bail!("Version to fetch does not match current resource.");
        }
    }

    let path = dest.as_ref().join("prs.json");
    let _ = fs::create_dir_all(dest);
    fs::write(&path, &bytes)?;

    Ok(())
}

#[tracing::instrument(skip_all)]
pub(crate) fn run(action: &Action) -> Result<()> {
    let config: Config = {
        let mut buf = String::new();
        io::stdin()
            .read_to_string(&mut buf)
            .wrap_err("Failed to read from stdin")?;

        if buf.is_empty() {
            eyre::bail!("No data received on stdin");
        }

        serde_json::from_str(&buf).wrap_err("Failed to parse stdin")?
    };

    match action {
        Action::Check => action_check(config),
        Action::In { dest } => action_in(config, dest),
        Action::Out => Ok(()),
    }
}
